using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using MinorShift.Emuera.Sub;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData.Variable;

namespace MinorShift.Emuera.GameData
{
	internal sealed class ConstantData
	{

		public const int MaxAbl = 100;
		public const int MaxTalent = 1000;
		public const int MaxExp = 100;
		public const int MaxMark = 100;
		public const int MaxTrain = 1000;
		public const int MaxParam = 200;
		public const int MaxItem = 1000;
		public const int MaxBasedata = 100;
		public const int MaxSource = 1000;
		public const int MaxEx = 1000;

		public const int MaxStr = 20000;
	
		public int[] VariableIntArrayLength;
		public int[] VariableStrArrayLength;
		public int[] CharacterIntArrayLength;
		public int[] CharacterStrArrayLength;

		public ConstantData(GameBase gamebase)
		{
			this.gamebase = gamebase;
			setDefaultArrayLength();
			AblName = new string[MaxAbl];
			BaseName = new string[MaxBasedata];
			TalentName = new string[MaxTalent];
			ExpName = new string[MaxExp];
			ParamName = new string[MaxParam];
			TrainName = new string[MaxTrain];
			MarkName = new string[MaxMark];
			SourceName = new string[MaxSource];
			ExName = new string[MaxEx];

			ItemName = new string[MaxItem];
			ItemPrice = new Int64[MaxItem];

			Str = new string[MaxStr];

			CharacterTmplList = new List<CharacterTemplate>();
		}

		private void setDefaultArrayLength()
		{
			VariableIntArrayLength = new int[(int)VariableCode.__COUNT_INTEGER_ARRAY__];
			VariableStrArrayLength = new int[(int)VariableCode.__COUNT_STRING_ARRAY__];
			CharacterIntArrayLength = new int[(int)VariableCode.__COUNT_CHARACTER_INTEGER_ARRAY__];
			CharacterStrArrayLength = new int[(int)VariableCode.__COUNT_CHARACTER_STRING_ARRAY__];

			for (int i = 0; i < VariableIntArrayLength.Length; i++)
				VariableIntArrayLength[i] = 1000;
			VariableIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.FLAG)] = 10000;
			VariableIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.ITEMPRICE)] = ConstantData.MaxItem;


			for (int i = 0; i < VariableStrArrayLength.Length; i++)
				VariableStrArrayLength[i] = 100;
			VariableStrArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.STR)] = ConstantData.MaxStr;

			for (int i = 0; i < CharacterIntArrayLength.Length; i++)
				CharacterIntArrayLength[i] = 100;
			CharacterIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.TALENT)] = 1000;
			CharacterIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.CFLAG)] = 1000;
			CharacterIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.JUEL)] = 200;
			CharacterIntArrayLength[(int)(VariableCode.__LOWERCASE__ & VariableCode.GOTJUEL)] = 200;

			for (int i = 0; i < CharacterStrArrayLength.Length; i++)
				CharacterStrArrayLength[i] = 100;

		}

		private void loadVariableSizeData(string csvPath)
		{
			if (!File.Exists(csvPath))
				return;
			EraStreamReader eReader = new EraStreamReader();
			if (!eReader.Open(csvPath))
			{
				output.PrintLine(eReader.Filename + "̃I[vɎs܂");
				return;
			}
			ScriptPosition position = null;
			output.PrintLine(eReader.Filename + "ǂݍݒEEE");
			try
			{
				string line = null;
				while ((line = eReader.ReadLine()) != null)
				{
					//ÕzCgXy[Xǂݎ̂
					StringStream st = new StringStream(line);
					TokenReader.SkipWhiteSpace(st);
					if ((st.EOS) || (st.Current == ';'))
						continue;
					line = st.Substring();
					position = new ScriptPosition(eReader.Filename, eReader.LineNo, line);
					string[] tokens = line.Split(',');
					if (tokens.Length < 2)
					{
						output.PrintWarning("\",\"Kvł", position, 1);
						continue;
					}
					string idtoken = tokens[0].Trim();
					VariableIdentifier id = VariableIdentifier.GetVariableId(idtoken);
					if (id == null)
					{
						output.PrintWarning("ڂ̒lϐƂĔFł܂", position, 1);
						continue;
					}
					int length = 0;
					if (!int.TryParse(tokens[1], out length))
					{
						output.PrintWarning("ڂ̒l𐮐lƂĔFł܂", position, 1);
						continue;
					}
					if (length < 100)
					{
						output.PrintWarning("zTCY100ȉɂ͏o܂", position, 1);
						continue;
					}
					if(!id.IsArray)
					{
						output.PrintWarning("zϐłȂϐ" + id.ToString() + "̃TCYύXł܂", position, 1);
						continue;
					}
					if ((id.IsCalc)&&(id.Code != VariableCode.LOCAL))
					{
						output.PrintWarning(id.ToString() + "̃TCY͕ύXł܂", position, 1);
						continue;
					}
					if (id.IsCharacterData)
					{
						if (id.IsInteger)
							CharacterIntArrayLength[id.CodeInt] = length;
						else if (id.IsString)
							CharacterStrArrayLength[id.CodeInt] = length;
					}
					else
					{
						if (id.IsInteger)
							VariableIntArrayLength[id.CodeInt] = length;
						else if (id.IsString)
							VariableStrArrayLength[id.CodeInt] = length;
					}
				}
			}
			catch
			{
				if (position != null)
					output.PrintWarning("\ȂG[܂", position, 3);
				else
					output.PrintLine("\ȂG[܂");
				return;
			}
			finally
			{
				eReader.Close();
			}


		}

		private readonly GameBase gamebase;
		private readonly string[][] names = new string[11][];
		public string[] AblName { get { return names[0]; } set { names[0] = value; }}
		public string[] ExpName { get { return names[1]; } set { names[1] = value; } }
		public string[] TalentName { get { return names[2]; } set { names[2] = value; } }
		public string[] ParamName { get { return names[3]; } set { names[3] = value; } }
		public string[] TrainName { get { return names[4]; } set { names[4] = value; } }
		public string[] MarkName { get { return names[5]; } set { names[5] = value; } }
		public string[] ItemName { get { return names[6]; } set { names[6] = value; } }
		public string[] BaseName { get { return names[7]; } set { names[7] = value; } }
		public string[] SourceName { get { return names[8]; } set { names[8] = value; } }
		public string[] ExName { get { return names[9]; } set { names[9] = value; } }
		public string[] Str { get { return names[10]; } set { names[10] = value; } }
		public Int64[] ItemPrice;

		private readonly List<CharacterTemplate> CharacterTmplList;

		private EmueraConsole output;
		public void LoadData(string csvDir, EmueraConsole console)
		{
			output = console;
			loadVariableSizeData(csvDir + "VariableSize.CSV");
			loadDataTo(csvDir + "ABL.CSV", AblName, null);
			loadDataTo(csvDir + "EXP.CSV", ExpName, null);
			loadDataTo(csvDir + "TALENT.CSV", TalentName, null);
			loadDataTo(csvDir + "PALAM.CSV", ParamName, null);
			loadDataTo(csvDir + "TRAIN.CSV", TrainName, null);
			loadDataTo(csvDir + "MARK.CSV", MarkName, null);
			loadDataTo(csvDir + "ITEM.CSV", ItemName, ItemPrice);
			loadDataTo(csvDir + "BASE.CSV", BaseName, null);
			loadDataTo(csvDir + "SOURCE.CSV", SourceName, null);
			loadDataTo(csvDir + "EX.CSV", ExName, null);
			loadDataTo(csvDir + "STR.CSV", Str, null);
			loadCharacterData(csvDir);
			definedNames.Clear();
			foreach (string[] strs in names)
				foreach (string str in strs)
					definedNames.Add(str);
			foreach (CharacterTemplate chara in CharacterTmplList)
			{
				definedNames.Add(chara.Name);
				definedNames.Add(chara.Callname);
			}
		}

		private static readonly List<string> definedNames = new List<string>();
		public static bool isDefined(string str)
		{
			if(string.IsNullOrEmpty(str))
				return false;
			return definedNames.Contains(str);
		}
		public int KeywordToInteger(VariableCode code, string key)
		{
			if(string.IsNullOrEmpty(key))
				throw new CodeEE("L[[hɂ͏o܂");
			int p1 = -1;
			string errPos = null;
			switch (code)
			{
				case VariableCode.ITEM:
				case VariableCode.ITEMSALES:
					p1 = ((IList<string>)ItemName).IndexOf(key);
					errPos = "Item.csv";
					break;
				case VariableCode.LOSEBASE:
				case VariableCode.BASE:
				case VariableCode.MAXBASE:
					p1 = ((IList<string>)BaseName).IndexOf(key);
					errPos = "base.csv";
					break;

				case VariableCode.ABL:
					p1 = ((IList<string>)AblName).IndexOf(key);
					errPos = "abl.csv";
					break;

				case VariableCode.TALENT:
					p1 = ((IList<string>)TalentName).IndexOf(key);
					errPos = "talent.csv";
					break;

				case VariableCode.EXP:
					p1 = ((IList<string>)ExpName).IndexOf(key);
					errPos = "exp.csv";
					break;
				case VariableCode.MARK:
					p1 = ((IList<string>)MarkName).IndexOf(key);
					errPos = "mark.csv";
					break;
				case VariableCode.SOURCE:
					p1 = ((IList<string>)SourceName).IndexOf(key);
					errPos = "source.csv";
					break;
				case VariableCode.EX:
				case VariableCode.NOWEX:
					p1 = ((IList<string>)ExName).IndexOf(key);
					errPos = "ex.csv";
					break;
				case VariableCode.RELATION:
					for (int i = 0; i < CharacterTmplList.Count; i++)
					{
						if ((CharacterTmplList[i].Name == key) || (CharacterTmplList[i].Callname == key))
						{
							p1 = (int)CharacterTmplList[i].No;
							break;
						}
					}
					errPos = "chara*.csv";
					break;


				case VariableCode.UP:
				case VariableCode.DOWN:
				case VariableCode.PALAM:
				case VariableCode.JUEL:
				case VariableCode.GOTJUEL:
					p1 = ((IList<string>)ParamName).IndexOf(key);
					errPos = "palam.csv";
					break;


				//case VariableCode.FLAG:
				//case VariableCode.TFLAG:
				//case VariableCode.CFLAG:
				//case VariableCode.EQUIP:
				//case VariableCode.TEQUIP:
				//case VariableCode.STAIN:
			}
			if (p1 < 0)
			{
				if (errPos == null)
					throw new CodeEE("zϐ" + code.ToString() + "̗vf𕶎Ŏw肷邱Ƃ͂ł܂");
				else
					throw new CodeEE(errPos + "̒\"" + key +"\"̒`܂");
			}
			return p1;
		}

		public CharacterTemplate GetCharacterTemplate(Int64 index, bool sp)
		{
			foreach (CharacterTemplate chara in CharacterTmplList)
			{
				if (chara.No != index)
					continue;
				if (sp == (chara.CFlag[0] != 0))
					return chara;
			}
			return null;
		}


		private void loadCharacterData(string csvDir)
		{
			if (!Directory.Exists(csvDir))
				return;
			string[] csvPaths = System.IO.Directory.GetFiles(csvDir, "CHARA*.CSV", SearchOption.TopDirectoryOnly);
			//NTFS̏ƃt@C͈Ⴄ炵
			//Array.Sort(csvPaths, new MinorShift.Emuera.GameProc.StrIgnoreCaseComparer());
			if (csvPaths.Length == 0)
				return;
			foreach (string path in csvPaths)
			{
				loadCharacterDataFile(path);
			}
		}

		private void loadCharacterDataFile(string csvPath)
		{
			CharacterTemplate tmpl = null;
			EraStreamReader eReader = new EraStreamReader();
			if (!eReader.Open(csvPath))
			{
				output.PrintLine(eReader.Filename + "̃I[vɎs܂");
				return;
			}
			ScriptPosition position = null;
			output.PrintLine(eReader.Filename + "ǂݍݒEEE");
			try
			{
				Int64 index = -1;
				string line = null;
				while ((line = eReader.ReadLine()) != null)
				{
					//ÕzCgXy[Xǂݎ̂
					StringStream st = new StringStream(line);
					TokenReader.SkipWhiteSpace(st);
					if ((st.EOS) || (st.Current == ';'))
						continue;
					line = st.Substring();
					string[] tokens = line.Split(',');
					position = new ScriptPosition(eReader.Filename, eReader.LineNo, line);
					if (tokens.Length < 2)
					{
						output.PrintWarning("\",\"Kvł", position, 1);
						continue;
					}
					if (tokens[0].Length == 0)
					{
						output.PrintWarning("\",\"Ŏn܂Ă܂", position, 1);
						continue;
					}
					if ((tokens[0].Equals("NO", StringComparison.CurrentCultureIgnoreCase))
						|| (tokens[0].Equals("ԍ", StringComparison.CurrentCultureIgnoreCase)))
					{
						if (tmpl != null)
						{
							output.PrintWarning("ԍdɒ`܂", position, 1);
							continue;
						}
						if (!Int64.TryParse(tokens[1], out index))
						{
							output.PrintWarning(tokens[1] + "𐮐lɕϊł܂", position, 1);
							continue;
						}
						tmpl = new CharacterTemplate(index, this);
						CharacterTmplList.Add(tmpl);
						continue;
					}
					if (tmpl == null)
					{
						output.PrintWarning("ԍ`Oɑ̃f[^n܂܂", position, 1);
						continue;
					}
					toCharacterTemplate(gamebase, position, tmpl, tokens);
				}
			}
			catch
			{
				if (position != null)
					output.PrintWarning("\ȂG[܂", position, 3);
				else
					output.PrintLine("\ȂG[܂");
				return;
			}
			finally
			{
				eReader.Dispose();
			}
		}

		private void toCharacterTemplate(GameBase gamebase,ScriptPosition position, CharacterTemplate chara, string[] tokens)
		{
			if (chara == null)
				return;
			Int64 p1 = 0;
			Int64 p2 = 0;
			bool p1isNumeric = Int64.TryParse(tokens[1], out p1);
			bool p2isNumeric = ((tokens.Length > 2) && (Int64.TryParse(tokens[2], out p2)));
			Int64[] array = null;
			switch (tokens[0].ToUpper())
			{
				case "NAME":
				case "O":
					chara.Name = tokens[1];
					return;
				case "CALLNAME":
				case "Ăі":
					chara.Callname = tokens[1];
					return;
				case "MARK":
				case "":
					array = chara.Mark;
					if (!p1isNumeric)
						p1 = ((IList<string>)MarkName).IndexOf(tokens[1]);
					break;
				case "EXP":
				case "o":
					array = chara.Exp;
					if (!p1isNumeric)
						p1 = ((IList<string>)ExpName).IndexOf(tokens[1]);
					break;
				case "ABL":
				case "\":
					array = chara.Abl;
					if (!p1isNumeric)
						p1 = ((IList<string>)AblName).IndexOf(tokens[1]);
					break;
				case "BASE":
				case "b":
					if (!p1isNumeric)
						p1 = ((IList<string>)BaseName).IndexOf(tokens[1]);
					array = chara.Maxbase;
					break;
				case "TALENT":
				case "f":
					array = chara.Talent;
					if (!p1isNumeric)
						p1 = ((IList<string>)TalentName).IndexOf(tokens[1]);
					break;
				case "RELATION":
				case "":
					array = chara.Relation;
					if (!p1isNumeric)
						break;
					break;
				case "CFLAG":
				case "tO":
					array = chara.CFlag;
					if (!p1isNumeric)
						break;
					break;
				default:
					output.PrintWarning("\"" + tokens[0] + "\"͉߂łȂʎqł", position, 1);
					return;
			}
			if (array == null)
			{
				output.PrintWarning("vO~X", position, 3);
				return;
			}
			if ((p1 < 0) || (p1 >= array.Length))
			{
				if (p1isNumeric)
					output.PrintWarning(p1.ToString() + "͔z͈̔͊Oł", position, 1);
				else if (tokens[1].Length == 0)
					output.PrintWarning("ڂ̎ʎq܂", position, 1);
				else
					output.PrintWarning("\"" + tokens[1] + "\"͉߂łȂʎqł", position, 1);
				return;
			}
			if (!p2isNumeric)
				p2 = 1;
			array[p1] = p2;
		}


		private void loadDataTo(string csvPath, string[] target, Int64[] targetI)
		{

			if (!File.Exists(csvPath))
				return;
			if (target == null)
				return;

			EraStreamReader eReader = new EraStreamReader();
			if (!eReader.Open(csvPath))
			{
				output.PrintLine(eReader.Filename + "̃I[vɎs܂");
				return;
			}
			ScriptPosition position = null;
			output.PrintLine(eReader.Filename + "ǂݍݒEEE");
			try
			{
				string line = null;
				while ((line = eReader.ReadLine()) != null)
				{
					//ÕzCgXy[Xǂݎ̂
					StringStream st = new StringStream(line);
					TokenReader.SkipWhiteSpace(st);
					if ((st.EOS) || (st.Current == ';'))
						continue;
					line = st.Substring();
					position = new ScriptPosition(eReader.Filename, eReader.LineNo, line);
					string[] tokens = line.Split(',');
					if (tokens.Length < 2)
					{
						output.PrintWarning("\",\"Kvł", position, 1);
						continue;
					}
					int index = 0;
					if (!Int32.TryParse(tokens[0], out index))
					{
						output.PrintWarning("ڂ̒l𐮐lɕϊł܂", position, 1);
						continue;
					}
					if ((index < 0) || (target.Length <= index))
					{
						output.PrintWarning(index.ToString() + "͔z͈̔͊Oł", position, 1);
						continue;
					}
					target[index] = tokens[1];
					if ((targetI != null) && (tokens.Length >= 3))
					{
						Int64 price;

						if (!Int64.TryParse(tokens[2], out price))
						{
							output.PrintWarning("zǂݎ܂", position, 1);
							continue;
						}

						targetI[index] = price;
					}
				}
			}
			catch
			{
				if (position != null)
					output.PrintWarning("\ȂG[܂", position, 3);
				else
					output.PrintLine("\ȂG[܂");
				return;
			}
			finally
			{
				eReader.Close();
			}


		}
	}

	internal sealed class CharacterTemplate
	{
		public CharacterTemplate(Int64 index, ConstantData constant)
		{
			int[] arraySize = constant.CharacterIntArrayLength;
			No = index;
			Abl = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.ABL)]];
			Mark = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.MARK)]];
			Exp = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.EXP)]];
			Talent = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.TALENT)]];
			Relation = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.RELATION)]];
			CFlag = new Int64[arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.CFLAG)]];

			int baseSize = arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.MAXBASE)];
			baseSize = Math.Min(baseSize, arraySize[(int)(VariableCode.__LOWERCASE__ & VariableCode.BASE)]);
			Maxbase = new Int64[baseSize];
		}
		public string Name;
		public string Callname;
		public readonly Int64 No;
		public readonly Int64[] Maxbase;
		public readonly Int64[] Mark;
		public readonly Int64[] Exp;
		public readonly Int64[] Abl;
		public readonly Int64[] Talent;
		public readonly Int64[] Relation;
		public readonly Int64[] CFlag;
	}
}
